//***************************************************************************//
//* Calculate cumulative distribution (needed for SPHERES PST simulation)   *//
//***************************************************************************//

//! Data transfer between frida and gsl_function.

typedef struct {
    POlc fc;
    int k;
    int j;
    double low;
    double hig;
} CumDisDatTyp;


namespace NCurveFile
{
    void DivideInterval( 
        vector<double> &y, const int ilow, const int ihig, CumDisDatTyp &data );
}

//! Calculate cumulative probability distribution on equidistant grid

/* Input:  `dis(x)`, a probability distribution.
   Output: `cum(r)`, the inverse of the cumulative probability distribution
     based on `dis`. It is implicitely understood that the abscissa of `cum`
     is an equidistant grid 2^m+1 points from 0 to 1.

   The resulting `cum` is used as follows. To draw random values x
   that are distributed according to `dis`:
   - Draw a random number r from a uniform distribution in the interval [0,1].
   - Calculate ilow=floor(r*2^m) and ihig = ilow+1.
   - Obtain x = cum(ilow) + (cum(ihig)-cum(ilow))*(r-ilow/2^m).
*/

void NCurveFile::CumulativeDistribution()
{
    NOlm::IterateC fiter;

    static int nb = 3;
    nb = iask( "How many twofold divisions", nb );
    if( nb<1 ) return;

    static double low=-3, hig=+3;
    low = dask( "Lower bound for Phi^-1(r)", low );
    hig = dask( "Upper bound for Phi^-1(r)", hig );
    if( low>=hig ) return;

    while( POlc fin = fiter() ){
        POld fout( new COld );
        fout->name = "Phi_" + fin->name;
        fout->xco = CCoord( "r", "" );
        fout->yco = CCoord( "Phi^-1(r)", "" );
        fout->ZCo = fin->ZCo;
        fout->lDoc = fin->lDoc;
        fout->lDoc.push_back( "cumulative distribution" );
        
        for( int j=0; j<fin->nJ(); j++ ){
            PCurve c = fin->VC(j);
            PSpec s( new CSpec );
            s->z = c->z;
            s->x.resize( (1<<nb)+1 );
            for( int i=0; i<(1<<nb)+1; ++i )
                s->x[i] = ((double)i)/(1<<nb);
            s->y.resize( (1<<nb)+1 );
            s->y[0] = low;
            s->y[1<<nb] = hig;

            // recursive solution (acts on S.y):
            CumDisDatTyp data;
            data.fc = fin;
            data.k = fiter.k();
            data.j = j;
            DivideInterval( s->y, 0, 1<<nb, data );

            fout->V.push_back(s);
        }

        NOlm::mem_store( fout );
    }
}


//! Calculate Phi^-1(r) by recursively dividing Phi-intervals.

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_roots.h>

double IntervalDifference( double mid, void *data )
{
    const CumDisDatTyp *mydata = (const CumDisDatTyp*) data;
    int j = mydata->j, k = mydata->k;
    POlc fc = mydata->fc;
    return NCurveFile::NumericIntegral( fc, k, j,
                                        mydata->low, mid )
        -  NCurveFile::NumericIntegral( fc, k, j,
                                        mid, mydata->hig );
}

void NCurveFile::DivideInterval(
    vector<double> &Y, const int ilow, const int ihig,
    CumDisDatTyp &data )
{
    if( ihig-ilow<=1 )
        return; // end of recursion

    // starting values for intersection:
    double low = Y[ilow], hig = Y[ihig], mid = (hig+low)/2;
    
    // integration limits for IntervalDifferences:
    data.low = low; data.hig = hig;

    // overhead for GSL root finder:
    const gsl_root_fsolver_type *root_solver = gsl_root_fsolver_brent;
    gsl_root_fsolver *s = gsl_root_fsolver_alloc( root_solver );
    gsl_function F;
    F.function = &IntervalDifference;
    F.params = &data;
    gsl_root_fsolver_set( s, &F, low, hig );
    int status, iter = 0, max_iter = 20;

    // iterative root finding:
    do {
        //printf( "%4d %4d %5d [%.7f, %.7f] %.7f\n",
        //        ilow, ihig, iter, low, hig, mid );
        iter++;
        status = gsl_root_fsolver_iterate (s);
        mid = gsl_root_fsolver_root (s);
        low = gsl_root_fsolver_x_lower (s);
        hig = gsl_root_fsolver_x_upper (s);
        status = gsl_root_test_interval( low, hig, 0, 1e-10 );
        // if (status == GSL_SUCCESS) printf ("Converged.\n");
    } while( status == GSL_CONTINUE && iter < max_iter );
    gsl_root_fsolver_free (s);
   
    // save result:
    int imid = (ilow+ihig)/2;
    Y[imid] = mid;

    // continue recursion:
    DivideInterval( Y, ilow, imid, data );
    DivideInterval( Y, imid, ihig, data );
}
