//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis                         
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-                                
//*  http://apps.jcns.fz-juelich.de/frida                                   
//**************************************************************************************************

//! \file  expr.hpp
//! \brief Expressions are parsed into CRef, CContext, CObj, virtual CNode.

#include <boost/shared_ptr.hpp>


//! The context in which a CNode shall be evaluated.

class CContext {
 public:
    int k;                         //!< Index of file we are refering to
    int j;                         //!< Index of scan we are refering to
    int i;                         //!< Index of point we are refering to
    const vector<double> *vt;      //!< A vector we may refer to
    double vt_step;                //!< Step of equidistant vt, otherwise 0
    enum TDim { _1, _VI, _VT } dim;//!< Kind of requested output
    bool want_error;               //!< Are errors requested?
    int nv;                        //!< Number of requested vector elements

    CContext( int _k=-1, int _j=-1, int _i=-1 );

    void request_VT( const vector<double> *_vt );
    void request_VI( int _nv );

    string context_info() const;
};


//! A tuple of nodes that yield indices k,j,i.

class CRef {
 public:
    RNode tk; //!< A node yielding the workspace index k.
    RNode tj; //!< A node yielding the spectrum index j.
    RNode ti; //!< A node yielding the point index i.

    CRef( RNode _tk=RNode(), RNode _tj=RNode(), RNode _ti=RNode() ) : tk(_tk), tj(_tj), ti(_ti) {}
    
    int get_k( const CContext& ctx ) const;
    int get_j( const CContext& ctx, int nj ) const;
    string ref_info() const;
};


//! Virtual base class for all kinds of nodes in an expression tree.

class CNode {
 public:
    virtual ~CNode() {};
    virtual void npar_exec( int *np ) const =0;
    //! Recursive: set *kd=true if a subnode is K dependent
    virtual void kdep_exec( bool *kd ) const {;} //!< default: do nothing

    int npar() const;
    bool k_dependent() const;
    virtual bool has_dummy() const =0; //!< does node depend on t ?
    virtual bool has_conv() const =0;  //!< does node depend on resolution ?

    //! Evaluate tree. Private. User calls pass through API implemented below.
    virtual const RObj tree_val( const CContext& ctx ) const =0;
    //! API for tree evaluation: compute one index.
    int tree_val_idx( const CContext& ctx, const string& name ) const;
    virtual bool looks_like_indices() const { return false; };
    RObjVecInt to_index_list( int siz ) const;
    
    //! API for tree evaluation: compute one numeric value.
    RObj tree_point_val( int k=-1, int j=-1, int i=-1 ) const;
    double tree_point_dbl( int k=-1, int j=-1, int i=-1 ) const;

    //! API for tree evaluation: compute an array.
    void tree_vec_val( vector<double> *ret, vector<double> *dret, int k=-1, int j=-1 ) const;

    virtual void set_coord( CCoord& ret, int k=-1 ) const =0;
    virtual string tree_info() const =0;  //!< introspection, for debugging

    static RObj eval( string expr );
};
