//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  node.hpp
//! \brief nodes within expression trees, inheriting from virtual base CNode

//! Base class CNode is declared elsewhere (expr.hpp).

typedef boost::shared_ptr<class CVar> PVar;


//! Node with a few arguments. Pure virtual base class.

class CNodeWithArgs: public CNode {
 protected:
    int narg;
    vector<RNode> arg;
 private:
    void npar_exec( int *np ) const;
 public:
    CNodeWithArgs( int _narg ) : narg(_narg) { arg.resize(narg); };
    bool has_dummy() const;
    bool has_conv() const;
    void kdep_exec( bool *kd ) const;
    virtual string tree_info() const =0;
};


//! Generic function/operator node. Pure virtual base class.

class CNodeFun: public CNodeWithArgs {
 protected:
    const class CFunc* fu;
 public:
    CNodeFun( int _narg, const class CFunc* _fu ) : CNodeWithArgs(_narg), fu(_fu) {};
    const RObj tree_val( const CContext& ctx ) const;
    virtual void set_coord( CCoord& ret, int k ) const =0;
    virtual string tree_info() const =0;
};


//! Function/operator node with one argument.

class CNodeFun1: public CNodeFun {
 public:
    CNodeFun1( const class CFunc *_fu, RNode a0 );
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
};


//! Function/operator node with two arguments.

class CNodeFun2: public CNodeFun {
 public:
    CNodeFun2( const class CFunc *_fu, RNode a0, RNode a1 );
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
};


//! Function/operator node with three arguments.

class CNodeFun3: public CNodeFun {
 public:
    CNodeFun3( const class CFunc *_fu, RNode a0, RNode a1, RNode a2 );
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
};


//! Range node.

class CNodeRange: public CNodeWithArgs {
 private:
    bool inclusive_end;
 public:
    CNodeRange( bool _inclusive_end, RNode _beg=RNode(), RNode _end=RNode(), RNode _step=RNode() )
        : CNodeWithArgs(3), inclusive_end(_inclusive_end)
        { arg[0] = _beg; arg[1] = _end; arg[2] = _step; };
    const RObj tree_val( const CContext& ctx ) const;
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
    bool looks_like_indices() const { return !arg[0] || arg[0]->looks_like_indices(); }
};


//! List node.

class CNodeList: public CNodeWithArgs {
 public:
    CNodeList( RNode _list, RNode _incr );
    vector<RNode> args() const { return arg; };
    const RObj tree_val( const CContext& ctx ) const;
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
    bool looks_like_indices() const { return arg[0] && arg[0]->looks_like_indices(); }
};


//! Generalized integral node.

class CNodeGeni: public CNodeWithArgs {
 private:
    const class CGeni* geni;
 public:
    CNodeGeni( const class CGeni *_geni, RRef _ref );
    CNodeGeni( const class CGeni *_geni, RNode a0 );
    CNodeGeni( const class CGeni *_geni, RNode a0, RNode a1 );
    const RObj tree_val( const CContext& ctx ) const;
    void set_coord( CCoord& ret, int k ) const;
    void kdep_exec( bool *kd ) const { *kd = true; }
    string tree_info() const;
};


//! Curve functional node.

class CNodeCvin: public CNodeWithArgs {
 private:
    const class CCvin* cvin;
    RRef ref;
 public:
    CNodeCvin( const class CCvin *_cvin, RRef _ref, RNode a0, RNode a1 );
    const RObj tree_val( const CContext& ctx ) const;
    void set_coord( CCoord& ret, int k ) const;
    void kdep_exec( bool *kd ) const { *kd = true; }
    string tree_info() const;
};


//! End node with fixed numeric value.

class CNodeVal: public CNode {
 private:
    RObj val;
    void npar_exec( int *np ) const {;}
 public:
    CNodeVal( RObj _val ) : val( _val ) {};
    CNodeVal( double _val );
    CNodeVal( int _val );
    const RObj tree_val( const CContext& ctx ) const;
    bool has_dummy() const { return false; }
    bool has_conv() const { return false; }
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
    bool looks_like_indices() const;
};



//! Indexed variable node.

class CNodeIva: public CNode {
 private:
    RVar var;
    RRef ref;
    void npar_exec( int *np ) const;
 public:
    CNodeIva( RVar _var, RRef _ref=RRef( new CRef() ) ) : var(_var), ref(_ref) {}
    const RObj tree_val( const CContext& ctx ) const;
    bool has_dummy() const { return false; }
    bool has_conv() const { return false; }
    void kdep_exec( bool *kd ) const { *kd = true; }
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const;
};


//! Identifier node.

class CNodeIdf: public CNode {
 private:
    string idf;
    void npar_exec( int *np ) const {;}
 public:
    CNodeIdf( string _s );
    const RObj tree_val( const CContext& ctx ) const;
    bool has_dummy() const { return false; }
    bool has_conv() const { return false; }
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const { return "$"; }
};


//! Function argument (t) node.

class CNodeDummy: public CNode {
 private:
    void npar_exec( int *np ) const {;}
 public:
    CNodeDummy() {}
    const RObj tree_val( const CContext& ctx ) const;
    bool has_dummy() const { return true; } // this node _is_ t, so it has t
    bool has_conv() const { return false; }
    void set_coord( CCoord& ret, int k ) const {
        throw "BUG: CNodeDummy::set_coord not expected to be needed"; }
    string tree_info() const { return "farg"; }
};


//! Curve evaluation node.

class CNodeCev: public CNode {
 private:
    RRef ref;
    RNode arg;
    void npar_exec( int *np ) const {;}
 public:
    CNodeCev( RRef& _ref, RNode _arg );
    const RObj tree_val( const CContext& ctx ) const;
    bool has_dummy() const { return arg->has_dummy(); }
    bool has_conv() const { return arg->has_conv(); }
    void kdep_exec( bool *kd ) const { *kd = true; }
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const { return "c[" + ref->ref_info() + "]"; }
};


//! Base class for mixing in data from a "convolution" file

class CNodeMixin: public CNode {
 protected:
    RNode shift;
 public:
    CNodeMixin( const RNode& _shift ) : shift(_shift) {}
    const RObj tree_val( const CContext& ctx ) const;
    virtual RObj copy_theory( const CContext& ctx, double theshift ) const =0;
    virtual RObjVecDbl convolve( const CContext& ctx, double theshift, const PSpec& sv,
                                 double conv_norm, double conv_step ) const =0;
    bool has_conv() const { return true; }
};

//! Base class for convoluting a function with a spectrum.

class CNodeConvBase: public CNodeMixin {
 protected:
    RNode theory;
    void npar_exec( int *np ) const;
 public:
    CNodeConvBase( const RNode& _theory, const RNode& _shift )
        : CNodeMixin( _shift ), theory(_theory) {}
    bool has_dummy() const { return theory->has_dummy(); }
    void kdep_exec( bool *kd ) const;
    void set_coord( CCoord& ret, int k ) const { theory->set_coord( ret, k ); }
};


//! Convolute a function with a spectrum.

class CNodeConv: public CNodeConvBase {
 public:
    CNodeConv( const RNode& _theory, const RNode& _shift=RNode( new CNodeVal( 0.0 ) ) )
        : CNodeConvBase( _theory, _shift ) {}
    RObj copy_theory( const CContext& ctx, double theshift ) const;
    RObjVecDbl convolve( const CContext& ctx, double theshift, const PSpec& sv,
                         double conv_norm, double conv_step ) const;
    string tree_info() const {
        return "conv(" + theory->tree_info() +","+ shift->tree_info() + ")"; }
};


//! Convolution a function with a spectrum, using the function's primitive.

class CNodePConv: public CNodeConvBase {
 public:
    CNodePConv( const RNode& _theory, const RNode& _shift=RNode( new CNodeVal( 0.0 ) ) )
        : CNodeConvBase( _theory, _shift ) {};
    RObj copy_theory( const CContext& ctx, double theshift ) const;
    RObjVecDbl convolve( const CContext& ctx, double theshift, const PSpec& sv,
                         double conv_norm, double conv_step ) const;
    string tree_info() const {
        return "pconv(" + theory->tree_info() +","+ shift->tree_info() + ")"; }
};


//! Dirac's delta function node, to request a copy of the resolution function.

class CNodeDirac: public CNodeMixin {
 private:
    void npar_exec( int *np ) const;
 public:
    CNodeDirac( const RNode& _shift=RNode( new CNodeVal( 0.0 ) ) )
        : CNodeMixin( _shift ) {};
    RObj copy_theory( const CContext& ctx, double theshift ) const;
    RObjVecDbl convolve( const CContext& ctx, double theshift, const PSpec& sv,
                         double conv_norm, double conv_step ) const;
    bool has_dummy() const { return true; } // has implicit t dependence
    bool has_conv() const { return true; }
    void kdep_exec( bool *kd ) const;
    void set_coord( CCoord& ret, int k ) const;
    string tree_info() const { return "resol(" + shift->tree_info() + ")"; }
};
